@@ -1,4 +1,5 @@
-# $Id: mp3info.rb 82 2008-04-19 10:59:02Z moumar $
+# coding:utf-8
+# $Id: mp3info.rb 98 2008-09-16 11:07:31Z moumar $
 # License:: Ruby
 # Author:: Guillaume Pierronnet (mailto:moumar_AT__rubyforge_DOT_org)
 # Website:: http://ruby-mp3info.rubyforge.org/
@@ -17,24 +18,33 @@ end
 
 class Mp3Info
 
-  VERSION = "0.6.5"
+  VERSION = "0.6.13"
 
   LAYER = [ nil, 3, 2, 1]
-  BITRATE = [
+  BITRATE = {
+    1 => 
     [
       [32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448],
       [32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384],
       [32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320] ],
+    2 => 
+    [
+      [32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256],
+      [8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160],
+      [8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160]
+    ],
+    2.5 => 
     [
       [32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256],
       [8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160],
       [8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160]
     ]
-  ]
-  SAMPLERATE = [
-    [ 44100, 48000, 32000 ],
-    [ 22050, 24000, 16000 ]
-  ]
+  }
+  SAMPLERATE = {
+    1 => [ 44100, 48000, 32000 ],
+    2 => [ 22050, 24000, 16000 ],
+    2.5 => [ 11025, 12000, 8000 ]
+  }
   CHANNEL_MODE = [ "Stereo", "JStereo", "Dual Channel", "Single Channel"]
 
   GENRES = [
@@ -67,7 +77,21 @@ class Mp3Info
 
   TAG1_SIZE = 128
   #MAX_FRAME_COUNT = 6  #number of frame to read for encoder detection
-  V1_V2_TAG_MAPPING = { 
+  
+  # map to fill the "universal" tag (#tag attribute)
+  # for id3v2.2
+  TAG_MAPPING_2_2 = { 
+    "title"    => "TT2",
+    "artist"   => "TP1", 
+    "album"    => "TAL",
+    "year"     => "TYE",
+    "tracknum" => "TRK",
+    "comments" => "COM",
+    "genre_s"  => "TCO"
+  }
+
+  # for id3v2.3 and 2.4
+  TAG_MAPPING_2_3 = { 
     "title"    => "TIT2",
     "artist"   => "TPE1", 
     "album"    => "TALB",
@@ -79,9 +103,10 @@ class Mp3Info
   
   # http://www.codeproject.com/audio/MPEGAudioInfo.asp
   SAMPLES_PER_FRAME = [
-    [384, 384, 384],    # Layer I   
-    [1152, 1152, 1152], # Layer II
-    [1152, 576, 576]    # Layer III
+    nil,
+    {1=>384, 2=>384, 2.5=>384},    # Layer I   
+    {1=>1152, 2=>1152, 2.5=>1152}, # Layer II
+    {1=>1152, 2=>576, 2.5=>576}    # Layer III
   ]
 
   # mpeg version = 1 or 2
@@ -102,8 +127,16 @@ class Mp3Info
   # variable bitrate => true or false
   attr_reader(:vbr)
 
-  # only used in vbr mode
-  attr_reader(:samples_per_frame)
+  # Hash representing values in the MP3 frame header. Keys are one of the following:
+  # - :private (boolean)
+  # - :copyright (boolean)
+  # - :original (boolean)
+  # - :padding (boolean)
+  # - :error_protection (boolean)
+  # - :mode_extension (integer in the 0..3 range)
+  # - :emphasis (integer in the 0..3 range)
+  # detailled explanation can be found here: http://www.mp3-tech.org/programmer/frame_header.html
+  attr_reader(:header)
 
   # length in seconds as a Float
   attr_reader(:length)
@@ -126,9 +159,6 @@ class Mp3Info
   # the original filename
   attr_reader(:filename)
 
-  # Moved hastag1? and hastag2? to be booleans
-  attr_reader(:hastag1, :hastag2)
-  
   # Test the presence of an id3v1 tag in file +filename+
   def self.hastag1?(filename)
     File.open(filename) { |f|
@@ -144,7 +174,6 @@ class Mp3Info
     }
   end
 
-
   # Remove id3v1 tag from +filename+
   def self.removetag1(filename)
     if self.hastag1?(filename)
@@ -160,81 +189,96 @@ class Mp3Info
     end
   end
 
-  # Instantiate Mp3Info object with name +filename+ and an
-  # options hash for ID3v2#new underlying object
-  def initialize(filename, id3v2_options = {})
-    $stderr.puts("#{self.class}::new() does not take block; use #{self.class}::open() instead") if block_given?
+  # Instantiate Mp3Info object with name +filename+.
+  # options hash is used for ID3v2#new. 
+  # You can specify :parse_tags => false to disable the processing 
+  # of the tags (read and write).
+  def initialize(filename, options = {})
+    warn("#{self.class}::new() does not take block; use #{self.class}::open() instead") if block_given?
     @filename = filename
-    @id3v2_options = id3v2_options
+    @tag_parsing_enabled = options.delete(:parse_tags)
+    if @tag_parsing_enabled == nil
+      @tag_parsing_enabled = true
+    end
+    @id3v2_options = options
     reload
   end
 
   # reload (or load for the first time) the file from disk
   def reload
     raise(Mp3InfoError, "empty file") unless File.size?(@filename)
+
+    @header = {}
     
-    @tag1 = {}
-    @tag1.extend(HashKeys)
-
-    @tag2 = ID3v2.new(@id3v2_options)
-
     @file = File.new(filename, "rb")
     @file.extend(Mp3FileMethods)
+    @tag1 = @tag = @tag1_orig = @tag_orig = {}
+    @tag1.extend(HashKeys)
+    @tag2 = ID3v2.new(@id3v2_options)
     
     begin
-      parse_tags
-      @tag1_orig = @tag1.dup
+      if @tag_parsing_enabled
+        parse_tags
+        @tag1_orig = @tag1.dup
 
-      @tag = {}
+        if hastag1?
+          @tag = @tag1.dup
+        end
 
-      if hastag1?
-	@tag = @tag1.dup
+        if hastag2?
+          @tag = {}
+          # creation of a sort of "universal" tag, regardless of the tag version
+          tag2_mapping = @tag2.version =~ /^2\.2/ ? TAG_MAPPING_2_2 : TAG_MAPPING_2_3
+          tag2_mapping.each do |key, tag2_name| 
+            tag_value = (@tag2[tag2_name].is_a?(Array) ? @tag2[tag2_name].first : @tag2[tag2_name])
+            next unless tag_value
+            @tag[key] = tag_value.is_a?(Array) ? tag_value.first : tag_value
+
+            if %w{year tracknum}.include?(key)
+              @tag[key] = tag_value.to_i
+            end
+            # this is a special case with id3v2.2, which uses
+            # old fashionned id3v1 genres
+            if tag2_name == "TCO" && tag_value =~ /^\((\d+)\)$/
+              @tag["genre_s"] = GENRES[$1.to_i]
+            end
+          end
+        end
+
+        @tag.extend(HashKeys)
+        @tag_orig = @tag.dup
       end
 
-      if hastag2?
-	@tag = {}
-      #creation of a sort of "universal" tag, regardless of the tag version
-	V1_V2_TAG_MAPPING.each do |key1, key2| 
-	  t2 = @tag2[key2]
-	  next unless t2
-	  @tag[key1] = t2.is_a?(Array) ? t2.first : t2
-
-	  if key1 == "tracknum"
-	    val = @tag2[key2].is_a?(Array) ? @tag2[key2].first : @tag2[key2]
-	    @tag[key1] = val.to_i
-	  end
-	end
-      end
-
-      @tag.extend(HashKeys)
-      @tag_orig = @tag.dup
-
 
       ### extracts MPEG info from MPEG header and stores it in the hash @mpeg
       ###  head (fixnum) = valid 4 byte MPEG header
       
       found = false
 
+      head = nil
       5.times do
 	head = find_next_frame() 
-	@mpeg_version = [2, 1][head[19]]
-	@layer = LAYER[bits(head, 18,17)]
-	next if @layer.nil?
-	@bitrate = BITRATE[@mpeg_version-1][@layer-1][bits(head, 15,12)-1]
-	@error_protection = head[16] == 0 ? true : false
-	@samplerate = SAMPLERATE[@mpeg_version-1][bits(head, 11,10)]
-	@padding = (head[9] == 1 ? true : false)
-	@channel_mode = CHANNEL_MODE[@channel_num = bits(head, 7,6)]
-	@copyright = (head[3] == 1 ? true : false)
-	@original = (head[2] == 1 ? true : false)
+        @first_frame_pos = @file.pos - 4
+        current_frame = Mp3Info.get_frames_infos(head)
+	@mpeg_version = current_frame[:mpeg_version]
+	@layer = current_frame[:layer]
+	@header[:error_protection] = head[16] == 0 ? true : false
+	@bitrate = current_frame[:bitrate]
+	@samplerate = current_frame[:samplerate]
+	@header[:padding] = current_frame[:padding]
+	@header[:private] = head[8] == 0 ? true : false
+	@channel_mode = CHANNEL_MODE[@channel_num = Mp3Info.bits(head, 7,6)]
+	@header[:mode_extension] = Mp3Info.bits(head, 5,4)
+	@header[:copyright] = (head[3] == 1 ? true : false)
+	@header[:original] = (head[2] == 1 ? true : false)
+	@header[:emphasis] = Mp3Info.bits(head, 1,0)
 	@vbr = false
 	found = true
-	break
+        break
       end
 
       raise(Mp3InfoError, "Cannot find good frame") unless found
 
-
       seek = @mpeg_version == 1 ? 
 	(@channel_num == 3 ? 17 : 32) :       
 	(@channel_num == 3 ?  9 : 17)
@@ -245,37 +289,33 @@ class Mp3Info
       if vbr_head == "Xing"
 	puts "Xing header (VBR) detected" if $DEBUG
 	flags = @file.get32bits
-	@streamsize = @frames = 0
-	flags[1] == 1 and @frames = @file.get32bits
-	flags[2] == 1 and @streamsize = @file.get32bits 
-	puts "#{@frames} frames" if $DEBUG
-	raise(Mp3InfoError, "bad VBR header") if @frames.zero?
+	stream_size = frame_count = 0
+	flags[1] == 1 and frame_count = @file.get32bits
+	flags[2] == 1 and stream_size = @file.get32bits 
+	puts "#{frame_count} frames" if $DEBUG
+	raise(Mp3InfoError, "bad VBR header") if frame_count.zero?
 	# currently this just skips the TOC entries if they're found
 	@file.seek(100, IO::SEEK_CUR) if flags[0] == 1
-	@vbr_quality = @file.get32bits if flags[3] == 1
+	#@vbr_quality = @file.get32bits if flags[3] == 1
 
-        @samples_per_frame = SAMPLES_PER_FRAME[@layer-1][@mpeg_version-1] 
-	@length = @frames * @samples_per_frame / Float(@samplerate)
+        samples_per_frame = SAMPLES_PER_FRAME[@layer][@mpeg_version] 
+	@length = frame_count * samples_per_frame / Float(@samplerate)
 
-	@bitrate = (((@streamsize/@frames)*@samplerate)/144) >> 10
+	@bitrate = (((stream_size/frame_count)*@samplerate)/144) / 1024
 	@vbr = true
       else
 	# for cbr, calculate duration with the given bitrate
-	@streamsize = @file.stat.size - (hastag1? ? TAG1_SIZE : 0) - (@tag2.valid? ? @tag2.io_position : 0)
-	@length = ((@streamsize << 3)/1000.0)/@bitrate
-	if @tag2["TLEN"]
-	  # but if another duration is given and it isn't close (within 5%)
-	  #  assume the mp3 is vbr and go with the given duration
-	  tlen = (@tag2["TLEN"].is_a?(Array) ? @tag2["TLEN"].last : @tag2["TLEN"]).to_i/1000
-	  percent_diff = ((@length.to_i-tlen)/tlen.to_f)
-	  if percent_diff.abs > 0.05
-	    # without the xing header, this is the best guess without reading
-	    #  every single frame
-	    @vbr = true
-	    @length = @tag2["TLEN"].to_i/1000
-	    @bitrate = (@streamsize / @bitrate) >> 10
-	  end
-	end
+	stream_size = @file.stat.size - (hastag1? ? TAG1_SIZE : 0) - (@tag2.io_position || 0)
+	@length = ((stream_size << 3)/1000.0)/@bitrate
+        # read the first 100 frames and decide if the mp3 is vbr and needs full scan
+        begin
+          bitrate, length = frame_scan(100)
+          if @bitrate != bitrate
+            @vbr = true
+            @bitrate, @length = frame_scan
+          end
+        rescue Mp3InfoInternalError
+        end
       end
     ensure
       @file.close
@@ -312,7 +352,7 @@ class Mp3Info
 
   # Does the file has an id3v1 or v2 tag?
   def hastag?
-    hastag1? or hastag2?
+    hastag1? || hastag2?
   end
 
   # Does the file has an id3v1 tag?
@@ -322,7 +362,7 @@ class Mp3Info
 
   # Does the file has an id3v2 tag?
   def hastag2?
-    @tag2.valid?
+    @tag2.parsed?
   end
 
   # write to another filename at close()
@@ -346,10 +386,18 @@ class Mp3Info
     end
     [pos, length]
   end
-  
+
+  # return the length in seconds of one frame
+  def frame_length
+    SAMPLES_PER_FRAME[@layer][@mpeg_version] / Float(@samplerate)
+  end
+
   # Flush pending modifications to tags and close the file
   def close
     puts "close" if $DEBUG
+    if !@tag_parsing_enabled
+      return 
+    end
     if @tag != @tag_orig
       puts "@tag has changed" if $DEBUG
 
@@ -359,9 +407,11 @@ class Mp3Info
 	  @tag1[k] = v
 	end
       end
-      
-      V1_V2_TAG_MAPPING.each do |key1, key2|
-        @tag2[key2] = @tag[key1] if @tag[key1]
+
+      # ruby-mp3info can only write v2.3 tags
+      TAG_MAPPING_2_3.each do |key, tag2_name|
+        @tag2.delete(TAG_MAPPING_2_2[key])
+        @tag2[tag2_name] = @tag[key] if @tag[key]
       end
     end
 
@@ -402,9 +452,8 @@ class Mp3Info
       raise(Mp3InfoError, "file is not writable") unless File.writable?(@filename)
       tempfile_name = nil
       File.open(@filename, 'rb+') do |file|
-	
 	#if tag2 already exists, seek to end of it
-	if @tag2.valid?
+	if @tag2.parsed?
 	  file.seek(@tag2.io_position)
 	end
   #      if @file.read(3) == "ID3"
@@ -439,15 +488,64 @@ class Mp3Info
 
   # inspect inside Mp3Info
   def to_s
-    s = "MPEG #{@mpeg_version} Layer #{@layer} #{@vbr ? "VBR" : "CBR"} #{@bitrate} Kbps #{@channel_mode} #{@samplerate} Hz length #{@length} sec. error protection #{@error_protection} "
+    s = "MPEG #{@mpeg_version} Layer #{@layer} #{@vbr ? "VBR" : "CBR"} #{@bitrate} Kbps #{@channel_mode} #{@samplerate} Hz length #{@length} sec. header #{@header.inspect} "
     s << "tag1: "+@tag1.inspect+"\n" if hastag1?
     s << "tag2: "+@tag2.inspect+"\n" if hastag2?
     s
   end
 
+  # iterates over each mpeg frame over the file, allowing you to 
+  # write some funny things, like an mpeg lossless cutter, or frame
+  # counter, or whatever you like ;) +frame+ is a hash with the following keys:
+  # :layer, :bitrate, :samplerate, :mpeg_version, :padding and :size (in bytes)
+  def each_frame
+    File.open(@filename, 'r') do |file|
+      file.seek(@first_frame_pos, File::SEEK_SET)
+      loop do
+        head = file.read(4).unpack("N").first
+        frame = Mp3Info.get_frames_infos(head)
+        file.seek(frame[:size] -4, File::SEEK_CUR)
+        yield frame
+        #puts "frame #{frame_count} len #{frame[:length]} br #{frame[:bitrate]} @file.pos #{@file.pos}"
+        break if file.eof?
+      end
+    end
+  end
 
 private
   
+  def Mp3Info.get_frames_infos(head)
+    # be sure we are in sync
+    if ((head & 0xffe00000) != 0xffe00000)    || # 11 bit MPEG frame sync
+       ((head & 0x00060000) == 0x00060000)    || #  2 bit layer type
+       ((head & 0x0000f000) == 0x0000f000)    || #  4 bit bitrate
+       ((head & 0x0000f000) == 0x00000000)    || #        free format bitstream
+       ((head & 0x00000c00) == 0x00000c00)    || #  2 bit frequency
+       ((head & 0xffff0000) == 0xfffe0000) 
+      raise Mp3InfoInternalError 
+    end
+    mpeg_version = [2.5, nil, 2, 1][bits(head, 20,19)]
+    
+    layer = LAYER[bits(head, 18,17)]
+    raise Mp3InfoInternalError if layer == nil || mpeg_version == nil
+
+    bitrate = BITRATE[mpeg_version][layer-1][bits(head, 15,12)-1]
+    samplerate = SAMPLERATE[mpeg_version][bits(head, 11,10)]
+    padding = (head[9] == 1)
+    if layer == 1
+      size = (12 * bitrate*1000.0 / samplerate + (padding ? 1 : 0))*4 
+    else # layer 2 and 3
+      size = 144 * (bitrate*1000.0 / samplerate) + (padding ? 1 : 0)
+    end
+    size = size.to_i
+    { :layer => layer,
+      :bitrate => bitrate,
+      :samplerate => samplerate,
+      :mpeg_version => mpeg_version,
+      :padding => padding,
+      :size => size }
+  end
+
   ### parses the id3 tags of the currently open @file
   def parse_tags
     return if @file.stat.size < TAG1_SIZE  # file is too small
@@ -465,7 +563,7 @@ private
     unless @tag1_parsed         # v1 tag at end
       # this preserves the file pos if tag2 found, since gettag2 leaves
       # the file at the best guess as to the first MPEG frame
-      pos = (@tag2.valid? ? @file.pos : 0)
+      pos = (@tag2.io_position || 0)
       # seek to where id3v1 tag should be
       @file.seek(-TAG1_SIZE, IO::SEEK_END) 
       if @file.read(3) == "TAG"
@@ -474,47 +572,23 @@ private
       @file.seek(pos)
     end
   end
-
-  ### reads in id3 field strings, stripping out non-printable chars
-  ###  len (fixnum) = number of chars in field
-  ### returns string
-  def read_id3_string(len)
-    #FIXME handle unicode strings
-    #return @file.read(len)
-    s = ""
-    len.times do
-      c = @file.getc
-      # only append printable characters
-      s << c if c >= 32 and c < 254
-    end
-    return s.strip
-    #return (s[0..2] == "eng" ? s[3..-1] : s)
-  end
   
   ### gets id3v1 tag information from @file
   ### assumes @file is pointing to char after "TAG" id
   def gettag1
     @tag1_parsed = true
-=begin
-    # FIXME remove that
-    pos = @file.pos
-    p @file.read
-    @file.seek(pos)
-    # FIXME remove that
-=end
-    @tag1["title"] = read_id3_string(30)
-    @tag1["artist"] = read_id3_string(30)
-    @tag1["album"] = read_id3_string(30)
-    year_t = read_id3_string(4).to_i
+    @tag1["title"] = @file.read(30).unpack("A*").first
+    @tag1["artist"] = @file.read(30).unpack("A*").first
+    @tag1["album"] = @file.read(30).unpack("A*").first
+    year_t = @file.read(4).to_i
     @tag1["year"] = year_t unless year_t == 0
     comments = @file.read(30)
-    if comments[-2] == 0
-      @tag1["tracknum"] = comments[-1].to_i
+    if comments.getbyte(-2) == 0
+      @tag1["tracknum"] = comments.getbyte(-1).to_i
       comments.chop! #remove the last char
     end
-    #@tag1["comments"] = comments.sub!(/\0.*$/, '')
-    @tag1["comments"] = comments.strip
-    @tag1["genre"] = @file.getc
+    @tag1["comments"] = comments.unpack("A*").first
+    @tag1["genre"] = @file.getbyte
     @tag1["genre_s"] = GENRES[@tag1["genre"]] || ""
 
     # clear empty tags
@@ -530,42 +604,47 @@ private
     # It should be at byte 0 when there's no id3v2 tag.
     # It should be at the end of the id3v2 tag or the zero padding if there
     #   is a id3v2 tag.
-
     #dummyproof = @file.stat.size - @file.pos => WAS TOO MUCH
     dummyproof = [ @file.stat.size - @file.pos, 2000000 ].min
     dummyproof.times do |i|
-      if @file.getc == 0xff
+      if @file.getbyte == 0xff
         data = @file.read(3)
         raise(Mp3InfoError, "end of file reached") if @file.eof?
-        head = 0xff000000 + (data[0] << 16) + (data[1] << 8) + data[2]
-        if check_head(head)
-            return head
-        else
-            @file.seek(-3, IO::SEEK_CUR)
+        head = 0xff000000 + (data.getbyte(0) << 16) + (data.getbyte(1) << 8) + data.getbyte(2)
+        begin
+          Mp3Info.get_frames_infos(head)
+          return head
+        rescue Mp3InfoInternalError
+          @file.seek(-3, IO::SEEK_CUR)
         end
       end
     end
-    raise Mp3InfoError, "cannot find a valid frame after reading #{dummyproof} bytes"
+    if @file.eof?
+      raise Mp3InfoError, "cannot find a valid frame: got EOF"
+    else
+      raise Mp3InfoError, "cannot find a valid frame after reading #{dummyproof} bytes"
+    end
   end
 
-  ### checks the given header to see if it is valid
-  ###  head (fixnum) = 4 byte value to test for MPEG header validity
-  ### returns true if valid, false if not
-  def check_head(head)
-    return false if head & 0xffe00000 != 0xffe00000    # 11 bit MPEG frame sync
-    return false if head & 0x00060000 == 0x00060000    #  2 bit layer type
-    return false if head & 0x0000f000 == 0x0000f000    #  4 bit bitrate
-    return false if head & 0x0000f000 == 0x00000000    #        free format bitstream
-    return false if head & 0x00000c00 == 0x00000c00    #  2 bit frequency
-    return false if head & 0xffff0000 == 0xfffe0000
-    true
+  def frame_scan(frame_limit = nil)
+    frame_count = bitrate_sum = 0
+    each_frame do |frame|
+      bitrate_sum += frame[:bitrate]
+      frame_count += 1
+      break if frame_limit && (frame_count >= frame_limit)
+    end
+    
+    average_bitrate = bitrate_sum/frame_count.to_f
+    length = (frame_count-1) * frame_length
+    [average_bitrate, length]
   end
 
+
   ### returns the selected bit range (b, a) as a number
   ### NOTE: b > a  if not, returns 0
-  def bits(n, b, a)
+  def self.bits(number, b, a)
     t = 0
-    b.downto(a) { |i| t += t + n[i] }
+    b.downto(a) { |i| t += t + number[i] }
     t
   end
 end
